from __future__ import annotations

import itertools
import json
import os
import secrets
import traceback
from abc import ABC, abstractmethod
from argparse import Namespace
from dataclasses import asdict, dataclass, field
from functools import lru_cache
from pathlib import Path
from typing import Any, Optional

import ibis
import ibis.expr.types as ir
import pandas as pd
import pyarrow as pa
from ibis import Schema
from ibis.backends import BaseBackend
from ibis.backends.duckdb import Backend as DuckDBBackend
from mpire import WorkerPool
from tqdm import tqdm

from vinyl.lib.db_methods import ValidationOutput
from vinyl.lib.dbt import DBTProject
from vinyl.lib.dbt_methods import DBTDialect, DBTVersion
from vinyl.lib.errors import VinylError, VinylErrorType
from vinyl.lib.utils.pkg import _get_project_directory
from vinyl.lib.utils.text import _extract_uri_scheme

_TEMP_PATH_PREFIX = "vinyl_"
_QUERY_LIMIT = 100000
_JOIN_STRING_HELPER = "_____"
_PARALLEL_DB_THREADS = 10

# NOTE: Starting in Ibis 10.0, Ibis now uses nomenclature for tables (i.e. catalog, database, name), that is distinct from how we usually think about it. Database, schema, name. Until we rename our variables, the variable naming will be confusing.


@dataclass
class SourceInfo:
    _name: str
    _location: str
    _schema: Schema | None
    _parent_resource: Any | None = None


class _ResourceConnector(ABC):
    """Base interface for handling connecting to resource and getting sources"""

    @abstractmethod
    def _list_sources(
        self, with_schema=False
    ) -> tuple[list[SourceInfo], list[VinylError]]:
        pass

    @abstractmethod
    def _connect(self) -> BaseBackend:
        pass

    @abstractmethod
    def sql_to_df(self, query: str, limit: int | None) -> pa.Table:
        pass

    @abstractmethod
    def validate_sql(self, query: str) -> ValidationOutput:
        pass

    def _get_name(self):
        return self.__class__.__name__.replace("Connector", "").lower()


# exists to help distinguish between table and database connectors
class _TableConnector(_ResourceConnector):
    _tbls: dict[str, ir.Table]
    _path: str

    def __init__(self, path: str):
        self._path = path

    def _get_table(self, path) -> ir.Table:
        # for file connectors, we reconnect to the individual file to get the correct table. Since these tables are not in memory, we need to read to get the location.
        adj_conn = self.__class__(path)
        adj_conn._connect()
        return next(iter(adj_conn._tbls.values()))

    def _generate_twin(self, path, sample_row_count=1000) -> ir.Table:
        tbl = self._get_table(path)
        row_count = tbl.count()
        sampled = tbl.filter(
            ibis.random() < ibis.least(sample_row_count / row_count, 1)
        )
        return sampled

    def sql_to_df(self, query: str, limit: int | None = _QUERY_LIMIT) -> pa.Table:
        pass

    def validate_sql(self, query: str) -> ValidationOutput:
        pass


class _DatabaseConnector(_ResourceConnector):
    _conn: BaseBackend
    _allows_multiple_databases: bool = True
    _allows_multiple_schemas: bool = True
    _tables: list[str]
    _excluded_dbs: list[str] = []
    _excluded_schemas: list[str] = []
    _excluded_tables: list[str] = []

    def _expand_table_names(self, location: str) -> list[str]:
        db, sch, table = location.split(_JOIN_STRING_HELPER)
        if table == "*":
            if (
                not self._allows_multiple_schemas
                and not self._allows_multiple_databases
            ):
                table_set = set(self._conn.list_tables())
            elif not self._allows_multiple_schemas:
                # likely won't work
                table_set = set(self._conn.list_tables(database=db))
            elif not self._allows_multiple_databases:
                table_set = set(self._conn.list_tables(database=sch))
            else:
                table_set = set(self._conn.list_tables(database=(db, sch)))

            adj_tables = list(table_set - set(self._excluded_tables))
        else:
            adj_tables = [table]

        adj_tables = [_JOIN_STRING_HELPER.join([db, sch, t]) for t in adj_tables]

        return adj_tables

    # Define the function that processes each item
    def _get_sources_for_table_names(
        self, pre: str, with_schema: bool
    ) -> tuple[SourceInfo | None, VinylError | None]:
        conn: BaseBackend = self._connect()
        db, sch, tbl = pre.split(_JOIN_STRING_HELPER)
        location = f"{db}.{sch}"
        full_location = f"{location}.{tbl}"

        try:
            if with_schema:
                ibis_table = None
                if (
                    not self._allows_multiple_databases
                    and not self._allows_multiple_schemas
                ):
                    ibis_table = conn.table(name=tbl)
                elif not self._allows_multiple_schemas:
                    ibis_table = conn.table(database=db, name=tbl)
                elif not self._allows_multiple_databases:
                    ibis_table = conn.table(database=sch, name=tbl)
                else:
                    ibis_table = conn.table(database=(db, sch), name=tbl)
            return SourceInfo(
                _name=tbl,
                _location=location,
                _schema=(ibis_table.schema() if with_schema else None),
            ), None
        except Exception as e:
            return None, VinylError(
                node_id=full_location,
                type=VinylErrorType.DATABASE_ERROR,
                dialect=None,
                msg=str(e),
                traceback=traceback.format_exc(),
            )

    def _find_sources_in_db(
        self,
        databases_override: list[str] | None = None,
        with_schema: bool = False,
    ) -> tuple[list[SourceInfo], list[VinylError]]:
        self._connect()

        expanded_table_names = []

        # get tables
        for loc in self._tables:
            split_ = loc.split(".")
            if len(split_) != 3:
                continue
            database, schema, table = split_
            if databases_override is not None:
                adj_databases = databases_override
            elif database == "*":
                if not hasattr(self._conn, "list_catalogs"):
                    raise ValueError(
                        f"Database specification required for this connector: {self.__class__.__name__}"
                    )
                adj_databases = list(
                    set(self._conn.list_catalogs()) - set(self._excluded_dbs)
                )
            else:
                adj_databases = [database]

            unadj_tbl_names = []
            for db in adj_databases:
                if not self._allows_multiple_schemas:
                    adj_schemas = ["main"]
                elif schema == "*":
                    schema_set = set(
                        self._conn.list_databases(catalog=db)
                        if self._allows_multiple_databases
                        else self._conn.list_databases()
                    )
                    adj_schemas = list(schema_set - set(self._excluded_schemas))
                else:
                    adj_schemas = [schema]

                for sch in adj_schemas:
                    unadj_tbl_names.append(_JOIN_STRING_HELPER.join([db, sch, table]))

        with WorkerPool(n_jobs=10, keep_alive=True) as pool:
            nested_table_names = list(
                tqdm(
                    pool.imap_unordered(self._expand_table_names, unadj_tbl_names),
                    total=len(unadj_tbl_names),
                    desc="Getting table names...",
                )
            )

        # flatten list
        expanded_table_names = list(itertools.chain(*nested_table_names))

        if with_schema:
            # make parallel calls to db
            with WorkerPool(n_jobs=10, keep_alive=True) as pool:
                results = list(
                    tqdm(
                        pool.imap_unordered(
                            self._get_sources_for_table_names,
                            [(pre, True) for pre in expanded_table_names],
                        ),
                        total=len(expanded_table_names),
                        desc="Getting table schemas if required...",
                    )
                )
        else:
            # no need to make parallel calls
            results = [
                self._get_sources_for_table_names(t, False)
                for t in expanded_table_names
            ]
        sources = [r[0] for r in results if r[0] is not None]
        errors = [r[1] for r in results if r[1] is not None]
        return sources, errors

    @classmethod
    def _create_twin_connection(cls, path: str) -> DuckDBBackend:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return ibis.duckdb.connect(path)

    @lru_cache()
    def _get_table(self, database: str, schema: str, table: str) -> ir.Table:
        conn = self._connect()
        if not self._allows_multiple_databases and not self._allows_multiple_schemas:
            return conn.table(table)
        elif not self._allows_multiple_schemas:
            # likely won't work
            return conn.table(database=database, name=table)
        elif not self._allows_multiple_databases:
            return conn.table(database=schema, name=table)

        return conn.table(database=(database, schema), name=table)

    def _generate_twin(
        self,
        twin_path: str,
        database: str,
        schema: str,
        table: str,
        sample_row_count: int | None = 1000,
    ) -> ir.Table:
        tbl = self._get_table(database, schema, table)
        if sample_row_count is None:
            sampled = tbl
        else:
            row_count = tbl.count()
            # using safer random() sample
            sampled = tbl.filter(
                ibis.random() < ibis.least(sample_row_count / row_count, 1)
            )
        pyarrow_table = sampled.to_pyarrow()
        conn = self._create_twin_connection(twin_path)
        # using raw sql to set the schema since the argument is not supported in the ibis api

        # create final table
        conn.raw_sql(f"CREATE SCHEMA IF NOT EXISTS {schema}")
        conn.raw_sql(f"USE {schema}")
        table_temp_name = f"table_{secrets.token_hex(8)}"
        temp_table = conn.create_table(
            table_temp_name, obj=pyarrow_table, overwrite=True
        )

        # reconnect to catch type errors
        reconn = ibis.duckdb.connect(twin_path)
        temp_table = reconn.table(table_temp_name, schema=schema)
        original_types = sampled.schema().types
        cast_dict = {}
        for i, (name, type) in enumerate(temp_table.schema().items()):
            cast_dict[name] = original_types[i]
        temp_table = temp_table.cast(cast_dict)

        # create final table and delete temp one
        final_table = conn.create_table(table, obj=temp_table, overwrite=True)
        conn.drop_table(table_temp_name)
        return final_table

    def sql_to_df(self, query: str, limit: int | None = _QUERY_LIMIT) -> pa.Table:
        pass

    def validate_sql(self, query: str) -> ValidationOutput:
        pass


class _FileConnector(_ResourceConnector):
    _conn: DuckDBBackend = ibis.duckdb.connect()
    _paths_visited: list[str] = []
    _excluded_dbs = ["system", "temp"]
    _excluded_schemas = ["information_schema", "pg_catalog"]
    _remote: bool
    _path: str

    def __init__(self, path: str):
        if scheme := _extract_uri_scheme(path):
            import fsspec

            self._conn.register_filesystem(fsspec.filesystem(scheme))
            self._remote = True
            self._path = path
        else:
            self._remote = False
            # adjust local path so it works even if you are not in the root directory
            self._path = os.path.join(_get_project_directory(), path)


class DatabaseFileConnector(_FileConnector, _DatabaseConnector):
    _excluded_tables: list[str] = ["sqlite_sequence"]
    _use_sqlite: bool

    def __init__(
        self, path: str, tables: list[str] = ["*.*.*"], use_sqlite: bool = False
    ):
        super().__init__(
            path
        )  # init method from _FileConnector, not Database Connector (because of ordering)
        self._tables = tables
        if any([len(t.split(".")) != 3 for t in tables]):
            raise ValueError(
                "tables must be a string of format 'database.schema.table'"
            )
        self._use_sqlite = use_sqlite
        if use_sqlite:
            self._allows_multiple_databases = False
            self._allows_multiple_schemas = False

    def _list_sources(
        self, with_schema=False
    ) -> tuple[list[SourceInfo], list[VinylError]]:
        out, errors = self._find_sources_in_db(
            with_schema=with_schema, databases_override=[self._database]
        )
        return out, errors

    def _connect(self) -> DuckDBBackend:
        if self._use_sqlite:
            self._conn = self._connect_helper_sqlite(self._path)
            return self._conn
        return self._connect_helper_duckdb(self._conn, self._path)

    @classmethod
    @lru_cache()
    def _connect_helper_duckdb(cls, conn: DuckDBBackend, path) -> DuckDBBackend:
        # caching ensures we don't attach a database from the same path twice
        name = cls._get_db_name(path)
        if path.endswith(".duckdb") or path.endswith(".db"):
            conn.attach(path, name, read_only=os.getenv("VINYL_READ_ONLY", False))

        elif path.endswith(".sqlite"):
            conn.attach(
                path, name, read_only=os.getenv("VINYL_READ_ONLY", False), sqlite=True
            )

        else:
            raise NotImplementedError(
                f"Connection for {path} not supported. Only .duckdb and .sqlite files are supported."
            )

        return conn

    @property
    def _database(self) -> str:
        return self._get_db_name(self._path)

    @classmethod
    @lru_cache()
    def _get_db_name(cls, path):
        name = Path(path).stem
        # handle situation where two database files have the same stem name
        if name in cls._conn.list_catalogs():
            name += str(secrets.token_hex(8))
        return name

    @staticmethod
    @lru_cache()
    def _connect_helper_sqlite(path: str):
        conn = ibis.sqlite.connect(path)
        return conn


class FileConnector(_FileConnector, _TableConnector):
    _tbls: dict[str, ir.Table]

    def __init__(self, path: str):
        super().__init__(
            path
        )  # init method from p_FileConnector, not Database Connector (because of ordering)
        self._tbls: dict[str, ir.Table] = {}

    def _connect(self) -> DuckDBBackend:
        # caching ensures we don't attach a database from the same path twice
        self._tbls = self._connect_helper(self._conn, self._path)
        return self._conn

    def _list_sources(
        self, with_schema=False
    ) -> tuple[list[SourceInfo], list[VinylError]]:
        self._connect()
        return [
            SourceInfo(
                _name=tbl.get_name(),
                _location=path,
                _schema=tbl.schema() if with_schema else None,
            )
            for path, tbl in self._tbls.items()
        ], None

    @classmethod
    @lru_cache()
    def _connect_helper(cls, conn, path) -> dict[str, ir.Table]:
        stem = Path(path).stem
        tbls = {}
        # caching ensures we don't attach a database from the same path twice
        if path.endswith(".csv"):
            tbls[path] = conn.read_csv(path, table_name=stem)
        elif path.endswith(".parquet"):
            tbls[path] = conn.read_parquet(path, table_name=stem)
        elif path.endswith(".json"):
            tbls[path] = conn.read_json(path, table_name=stem)
        elif os.path.isdir(path):
            for sub in os.listdir(path):
                path_it = os.path.join(path, sub)
                # only looking at files prevents recursion into subdirectories
                if os.path.isfile(path_it) and not sub.startswith("."):
                    tbls.update(cls._connect_helper(conn, path_it))

        else:
            raise NotImplementedError(
                f"Connection for {path} not supported. Only .csv, .parquet, and .json files are supported"
            )

        return tbls


class BigQueryConnector(_DatabaseConnector):
    _credentials: Any
    _BQ_ITERATOR_ROW_CUTOFF = 1000  # faster to use iterators for small datasets, but faster to download directly for large datasets

    def __init__(
        self,
        tables: list[str],
        service_account_path: str | None = None,
        service_account_info: dict[str, str] | None = None,
    ):
        self._service_account_path = service_account_path
        self._service_account_info = service_account_info
        self._tables = tables

    def _connect(self) -> BaseBackend:
        self._conn = BigQueryConnector._connect_helper(
            self._service_account_path,
            json.dumps(self._service_account_info)
            if self._service_account_info
            else None,
        )
        return self._conn

    def _list_sources(self, with_schema=False) -> list[SourceInfo]:
        conn = self._connect()
        out, errors = self._find_sources_in_db(
            with_schema=with_schema,
            databases_override=[
                self._conn.client.project
            ],  # bigquery only allows on project per connection
        )
        return out, errors

    # caching ensures we create one bq connection per set of credentials across instances of the class
    @staticmethod
    @lru_cache()
    def _connect_helper(
        service_account_path: str | None = None,
        service_account_info_str: str | None = None,
    ) -> BaseBackend:
        from google.oauth2 import service_account

        if service_account_path is not None:
            credentials = service_account.Credentials.from_service_account_file(
                service_account_path
            )
        elif service_account_info_str is not None:
            if isinstance(service_account_info_str, str):
                service_account_info = json.loads(service_account_info_str)
            credentials = service_account.Credentials.from_service_account_info(
                service_account_info
            )
        else:
            credentials = None
        conn = ibis.bigquery.connect(credentials=credentials)

        # fixes BQ-specific error related to caching
        if not hasattr(conn, "_session_dataset") and isinstance(
            conn, ibis.backends.bigquery.Backend
        ):
            conn._session_dataset = None

        return conn

    def sql_to_df(self, query: str, limit: int | None = _QUERY_LIMIT) -> pa.Table:
        conn = self._connect()
        if limit is not None:
            query = f"select * from ({query}) limit {limit}"
        rows = conn.client.query_and_wait(query)
        if rows.total_rows < self._BQ_ITERATOR_ROW_CUTOFF:
            # faster to use iterators for small datasets
            data = [dict(row) for row in rows]
            columns = [field.name for field in rows.schema]
            df = pd.DataFrame(data, columns=columns)
        else:
            # faster to download directly for large datasets
            df = rows.to_dataframe()

        return df

    def validate_sql(self, query: str) -> ValidationOutput:
        from google.cloud.bigquery import QueryJobConfig

        conn = self._connect()

        job_config = QueryJobConfig(dry_run=True, use_query_cache=False)
        try:
            query_job = conn.client.query(query, job_config=job_config)
            bytes_processed = query_job.total_bytes_processed
            cost = bytes_processed * 5 * 1e-12  # relatively accurate for BigQuery
            return ValidationOutput(
                errors=None, bytes_processed=bytes_processed, cost=cost
            )
        except Exception as e:
            msg = str(e).split("\n\n")[0].split("prettyPrint=false: ")[-1]
            error_parsed = VinylError(
                "NA", VinylErrorType.DATABASE_ERROR, msg, dialect=self._get_name()
            )
            return ValidationOutput(
                errors=[error_parsed], bytes_processed=None, cost=None
            )


class PostgresConnector(_DatabaseConnector):
    _excluded_schemas = [
        "information_schema",
        "pg_catalog",
        "pgsodium",
        "auth",
        "extensions",
        "net",
    ]
    _allows_multiple_databases: bool = False
    _host: str
    _port: int
    _user: str
    _password: str
    _database: str

    def __init__(
        self,
        host: str,
        port: int,
        user: str,
        password: str,
        tables: list[str],
    ):
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._tables = tables

        # postgres requires connecting at the database level
        dbs = set([t.split(".")[0] for t in tables])
        if len(dbs) > 1 or "*" in dbs:
            raise ValueError("Postgres connector only supports one database at a time")
        self._database = dbs.pop()

    def _connect(self) -> BaseBackend:
        self._conn = self._connect_helper(
            self._host, self._port, self._user, self._password, self._database
        )
        return self._conn

    def _list_sources(self, with_schema=False) -> list[SourceInfo]:
        self._connect()
        out, errors = self._find_sources_in_db(with_schema=with_schema)
        return out, errors

    # caching ensures we create one bq connection per set of credentials across instances of the class
    @staticmethod
    @lru_cache()
    def _connect_helper(
        host: str, port: int, user: str, password: str, database: str
    ) -> BaseBackend:
        return ibis.postgres.connect(
            host=host, port=port, user=user, password=password, database=database
        )


class SnowflakeConnector(_DatabaseConnector):
    _account: str
    _user: str
    _password: str
    _warehouse: str | None

    def __init__(
        self,
        account: str,
        user: str,
        password: str,
        tables: list[str],
        warehouse: str | None = None,
    ):
        self._account = account
        self._user = user
        self._password = password
        self._warehouse = warehouse
        self._tables = tables

    def _connect(self) -> BaseBackend:
        self._conn = self._connect_helper(
            self._account, self._user, self._password, self._warehouse
        )
        return self._conn

    def _list_sources(self, with_schema=False) -> list[SourceInfo]:
        self._connect()
        out, errors = self._find_sources_in_db(with_schema=with_schema)
        return out, errors

    # caching ensures we create one bq connection per set of credentials across instances of the class
    @staticmethod
    @lru_cache()
    def _connect_helper(
        account: str, user: str, password: str, warehouse: str | None
    ) -> BaseBackend:
        if warehouse is not None:
            return ibis.snowflake.connect(
                account=account, user=user, password=password, warehouse=warehouse
            )
        else:
            return ibis.snowflake.connect(account=account, user=user, password=password)

    def sql_to_df(self, query: str, limit: int | None = _QUERY_LIMIT) -> pa.Table:
        conn = self._connect()
        if limit is not None:
            query = f"select * from ({query}) limit {limit}"
        cursor = conn.raw_sql(query)
        return cursor.fetch_pandas_all()

    def validate_sql(self, query: str) -> ValidationOutput:
        conn = self._connect()
        query = f"EXPLAIN USING JSON ({query})"
        try:
            cursor = conn.raw_sql(query)
            out = json.loads(cursor.fetchall()[0][0])
            bytes_processed = out["GlobalStats"]["bytesAssigned"]
            cost = (
                bytes_processed * 5 * 1e-12
            )  # guess based on BQ pricing, need to update
            return ValidationOutput(
                errors=None, bytes_processed=bytes_processed, cost=cost
            )
        except Exception as e:
            error = VinylError(
                "NA",
                VinylErrorType.DATABASE_ERROR,
                str(e).split(": ", 1)[-1],
                dialect=self._get_name(),
            )
            return ValidationOutput(errors=[error], bytes_processed=None, cost=None)


@dataclass(frozen=True)
class DBTArgs:
    profiles_dir: str
    project_dir: str
    profile: Optional[str] = None
    target: Optional[str] = None
    vars: dict[str, Any] = field(default_factory=dict)
    threads: Optional[int] = None

    def to_namespace(self) -> Namespace:
        self_as_dict = asdict(self)
        return Namespace(**self_as_dict)


class DBTConnectorFull:
    def __init__(
        self,
        dbt_project_dir: str,
        dialect: DBTDialect,
        version: DBTVersion,
        dbt_profiles_dir: str | None = None,
        tables: list[str] = [],
    ):
        """
        For dbt_profiles_dir, please provide the folder (e.g. "~/.dbt"), not the file (e.g. `~/.dbt/profiles.yml`)
        """
        self.dbt_project_dir = dbt_project_dir
        self.dialect = dialect
        self.version = version
        self.tables = tables

    def __new__(
        cls,
        dbt_project_dir: str,
        dialect: DBTDialect,
        version: DBTVersion,
        dbt_profiles_dir: str | None = None,
        tables: list[str] = ["*.*.*"],
    ):
        credentials = cls._get_dbt_credentials(
            dbt_project_dir, dialect, version, dbt_profiles_dir=dbt_profiles_dir
        )
        credentials_type = type(credentials).__name__
        # if tables == []:
        #     tables = [f"{credentials.database}.{credentials.schema}.*"]
        if credentials_type == "BigQueryCredentials":
            if str(credentials.method) == "service-account":
                return BigQueryConnector(
                    service_account_path=credentials.keyfile, tables=tables
                )
            elif str(credentials.method == "service-account-json"):
                return BigQueryConnector(
                    service_account_info=credentials.keyfile_json, tables=tables
                )
            else:
                # using oauth
                return BigQueryConnector(tables=tables)
        elif credentials_type == "SnowflakeCredentials":
            pass

        elif credentials_type == "PostgresCredentials":
            return PostgresConnector(
                host=credentials.host,
                port=credentials.port,
                user=credentials.user,
                password=credentials.password,
                tables=tables,
            )

        elif credentials_type == "DuckDBCredentials":
            return DatabaseFileConnector(path=credentials.path, tables=tables)

        else:
            raise ValueError(
                f"Dialect {credentials_type.split('Credentials')[0]} is not supported"
            )

    @lru_cache()
    @staticmethod
    def _get_dbt_credentials(
        dbt_project_dir: str,
        dialect: DBTDialect,
        version: DBTVersion,
        dbt_profiles_dir: str,
    ):
        dbtproj = DBTProject(
            dbt_project_dir, dialect, version, dbt_profiles_dir=dbt_profiles_dir
        )
        credentials = dbtproj.get_profile_credentials(via_dbt=False)
        return credentials


class DBTConnectorLite:
    def __init__(
        self,
        dialect: DBTDialect,
        profile_contents: dict[str, Any],
        tables: list[str] = ["*.*.*"],
    ):
        self.dialect = dialect
        self.profile_contents = profile_contents
        self.tables = tables

    def __new__(
        cls,
        dialect: DBTDialect,
        profile_contents: dict[str, Any],
        tables: list[str] = ["*.*.*"],
    ):
        target = profile_contents["target"]
        output = profile_contents["outputs"][target]
        if dialect == DBTDialect.BIGQUERY:
            if output["method"] == "service-account-json":
                return BigQueryConnector(
                    service_account_info=output["keyfile_json"], tables=tables
                )
            elif output["method"] == "service-account":
                return BigQueryConnector(
                    service_account_path=output["keyfile"], tables=tables
                )
        elif dialect == DBTDialect.SNOWFLAKE:
            return SnowflakeConnector(
                account=output["account"],
                user=output["user"],
                password=output["password"],
                warehouse=output.get("warehouse", None),
                tables=tables,
            )
        else:
            raise ValueError(f"Dialect {dialect} is not supported")



from vinyl.lib.project import Project
from vinyl.lib.table import VinylTable


class QueryEngineResult:
    def __init__(self, table: VinylTable):
        self.table = table

    def to_pandas(self):
        return self.table.execute("pandas")

    def to_records(self):
        df = self.to_pandas()
        for col, dtype in df.dtypes.items():
            if dtype.kind == "M":  # Check for datetime64 types
                df[col] = df[col].dt.strftime("%Y-%m-%dT%H:%M:%S")

        return df.to_dict("records")

    def columns(self):
        return self.table.columns

    def numpy_rows(self):
        df = self.to_pandas()
        return [tuple(row) for row in df.to_numpy()]


class QueryEngine:
    def __init__(self, project: Project):
        self.project = project

    def _model(self, name: str, limit: int):
        table = self.project._get_model(name)
        return QueryEngineResult(table.limit(limit))

    def _metric(
        self,
        store: str,
        grain: str,
        metrics: list[str] | None = None,
        dimensions: list[str] | None = None,
        limit: int | None = 1000,
    ):
        if "=" not in grain:
            raise ValueError("Invalid bucket format. Ex: 'days=1'")

        bucket_grain, bucket_value = grain.split("=")
        met_store = self.project._get_metric_store(store)
        combined_keys = [k for k in met_store._metrics_dict] + [
            k for k in met_store._derived_metrics
        ]

        cols = [
            getattr(met_store, name)
            for name in combined_keys
            if metrics is None or name in metrics
        ]

        if dimensions is None:
            dimensions = []
        else:
            dimensions = [getattr(met_store, name) for name in dimensions]

        table = met_store.select(
            [
                met_store.ts.dt.floor(**{bucket_grain: int(bucket_value)}),
                *dimensions,
                *cols,
            ]
        )

        if limit is None:
            return QueryEngineResult(table)

        else:
            return QueryEngineResult(table.limit(limit))



import re
from urllib.parse import urlparse


def parse_urn(urn):
    # Use regular expressions to extract the main URN and the nested URN
    pattern = r"^(urn:[^:]+:[^:]+):\((urn:[^:]+:[^:]+:[^:]+),([^,]+),([^)]+)\)$"
    match = re.match(pattern, urn)

    if not match:
        raise ValueError("Invalid URN format")

    main_urn = match.group(1)
    nested_urn = match.group(2)
    identifier = match.group(3)
    environment = match.group(4)

    return {
        "main_urn": main_urn,
        "nested_urn": parse_nested_urn(nested_urn),
        "identifier": identifier,
        "environment": environment,
    }


def parse_nested_urn(urn):
    # Parse the nested URN
    parsed_urn = urlparse(urn)

    scheme = parsed_urn.scheme
    namespace = parsed_urn.netloc
    entity_type = parsed_urn.path.split(":")[0]
    entity_name = parsed_urn.path.split(":")[1]

    return {
        "scheme": scheme,
        "namespace": namespace,
        "entity_type": entity_type,
        "entity_name": entity_name,
    }


# Example URN with nested URN
urn = "urn:li:dataset:(urn:li:dataPlatform:bigquery,capchase.analytics.parameters,PROD)"

parsed_urn = parse_urn(urn)

print(parsed_urn)



import pytest
from fastapi.testclient import TestClient

from vinyl.infra.http_server.server import (
    app,
    get_project,
)
from vinyl.lib.project import Project


@pytest.fixture(scope="module")
def client():
    app.dependency_overrides[get_project] = lambda: Project.bootstrap()
    client = TestClient(app)
    return client


def test_root(client):
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"status": "OK"}


def test_query_model(client):
    response = client.get("/models/amount_base?limit=10")
    assert response.status_code == 200


def test_query_metric(client):
    response = client.get("/metrics/fare_metrics?grain=months=1&limit=10")
    assert response.status_code == 200



import glob
import os

import toml


def delete_pyi_files(directory):
    # Check if the given directory exists
    if not os.path.exists(directory):
        print(f"The directory '{directory}' does not exist.")
        return

    # Use glob.glob with recursive=True to find all .pyi files in the directory and its subdirectories
    pyi_files = glob.glob(os.path.join(directory, "**/*.pyi"), recursive=True)

    # Loop through the list of .pyi files and remove each one
    for file in pyi_files:
        try:
            os.remove(file)
            print(f"Deleted: {file}")
        except OSError as e:
            print(f"Error deleting {file}: {e.strerror}")


# Replace 'your_directory_path_here' with the actual directory path

if __name__ == "__main__":
    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)
        project_name = pyproject["tool"]["poetry"]["name"]
    delete_pyi_files(project_name)



## NOTE: can't name this file `dbt.py` or weird import errors will ensue.
import os
import re
import subprocess
import sys
import traceback
from dataclasses import make_dataclass
from io import StringIO
from typing import Any

import yaml
from dbt_extractor import ExtractionError, py_extract_from_source

from vinyl.lib.dbt_methods import (
    MAX_DIALECT_VERSION,
    DBTArgs,
    DBTDialect,
    DBTError,
    DBTVersion,
)
from vinyl.lib.errors import VinylError, VinylErrorType
from vinyl.lib.utils.files import adjust_path, cd, file_exists_in_directory, load_orjson
from vinyl.lib.utils.graph import DAG


class DBTProject(object):
    compiled_sql_path: str
    manifest_path: str
    catalog_path: str
    target_path: str
    manifest: dict[str, Any]
    model_graph: DAG
    catalog: dict[str, Any]
    profile_name: str
    dbt_profiles_dir: str
    dbt_project_yml: dict[str, Any]
    dbt1_5: bool
    max_threads: int = 4
    version: DBTVersion
    version_list: list[int]

    def __init__(
        self,
        dbt_project_dir: str,
        dialect: DBTDialect,
        version: DBTVersion,
        dbt_profiles_dir: str | None = None,
        target_path: str | None = None,
        manifest_req: bool = False,
        catalog_req: bool = False,
        compile_exclusions: list[str] | None = None,
        env_vars: dict[str, str] = {},
    ):
        self.dbt_project_dir = adjust_path(dbt_project_dir)
        self.dialect = dialect
        self.version = version
        self.compile_exclusions = compile_exclusions
        self.env_vars = env_vars

        # set version bool
        self.version_list = [int(v[0]) for v in self.version.value.split(".")]
        if tuple(self.version_list) >= (1, 5):
            self.dbt1_5 = True
        else:
            self.dbt1_5 = False

        # get target paths
        if target_path:
            self.target_path = adjust_path(target_path)
            os.environ["DBT_TARGET_PATH"] = self.target_path
        else:
            self.get_project_yml_files()
        self.generate_target_paths()

        # get profiles-dir
        self.dbt_profiles_dir = (
            self.get_profiles_dir() if not dbt_profiles_dir else dbt_profiles_dir
        )

        if self.needs_dbt_deps():
            self.run_dbt_command(["deps"])

        # get artifacts if required
        if manifest_req:
            self.mount_manifest()

        if catalog_req:
            self.mount_catalog()

        # validate paths
        if not os.path.exists(self.dbt_project_dir):
            raise Exception(f"Project path do not exist: {self.dbt_project_dir}")

        self.install_dbt_if_necessary()

    def install_dbt_if_necessary(self):
        dbt_package = f"dbt-{self.dialect.value.lower()}"
        install_package = f"{dbt_package}~={self.version.value}.0"
        res = subprocess.run(["uv", "pip", "show", dbt_package, "--python", sys.executable], capture_output=True)
        if res.returncode != 0:
            subprocess.run(
                [
                    "uv",
                    "pip",
                    "install",
                    install_package,
                    "--python",
                    sys.executable,
                ],
                check=True,
            )
            return

        major_version_in_stdout = (
            res.stdout.decode().split("\n")[1].split(":")[1].strip().rsplit(".", 1)[0]
        )
        if major_version_in_stdout != self.version.value:
            subprocess.run(
                [
                    "uv",
                    "pip",
                    "install",
                    install_package,
                ],
                check=True,
            )

    def generate_target_paths(self):
        if self.target_path:
            self.manifest_path = os.path.join(self.target_path, "manifest.json")
            self.catalog_path = os.path.join(self.target_path, "catalog.json")
            self.compiled_sql_path = os.path.join(self.target_path, "compiled")
        else:
            raise Exception("Target path not found")

    def needs_dbt_deps(self):
        packages_yml_path = os.path.join(self.dbt_project_dir, "packages.yml")
        dbt_packages_path = os.path.join(self.dbt_project_dir, "dbt_packages")
        if not os.path.exists(packages_yml_path):
            return False
        if os.path.exists(dbt_packages_path) and os.listdir(dbt_packages_path):
            return False
        return True

    def mount_manifest(self, read=True, force_read=False, force_run=False):
        if hasattr(self, "manifest") and not force_run and not force_read:
            return
        elif force_run or not os.path.exists(self.manifest_path):
            self.dbt_parse()

        if read:
            self.manifest = load_orjson(self.manifest_path)
        elif (force_read or force_run) and hasattr(self, "manifest"):
            # make sure you don't accidentally read an outdated manifest in the future
            del self.manifest

    def build_model_graph(self, include_sources: bool = False, rebuild=False) -> DAG:
        if hasattr(self, "model_graph") and not rebuild:
            return self.model_graph
        # build networkx graph
        dag = DAG()
        for parent, children in self.manifest["child_map"].items():
            for child in children:
                dag.add_edge(parent, child)

        # build table_level dag without metric nodes
        nodes_to_remove = [
            k
            for k in dag.node_dict
            if k.startswith("metric.")
            or k.startswith("test.")
            or (
                not include_sources
                and (k.startswith("source.") or k.startswith("seed."))
            )  # only include sources and seeds in nodes to remove if include_sources is True
        ]
        dag.remove_nodes_and_reconnect(nodes_to_remove)

        self.model_graph = dag
        return dag

    def get_ancestors_and_descendants(
        self,
        node_ids: list[str],
        predecessor_depth: int | None,
        successor_depth: int | None,
    ) -> list[str]:
        self.build_model_graph(include_sources=False)
        return self.model_graph.get_ancestors_and_descendants(
            node_ids, predecessor_depth, successor_depth
        )

    def mount_catalog(self, read=True, force_run=False, force_read=False):
        if hasattr(self, "catalog") and not force_run and not force_read:
            return
        elif force_run or not os.path.exists(self.catalog_path):
            self.dbt_docs_generate()

        if read:
            self.catalog = load_orjson(self.catalog_path)
        elif (force_read or force_run) and hasattr(self, "catalog"):
            # make sure you don't accidentally read an outdated artifact in the future
            del self.catalog

    def get_project_yml_files(self):
        with open(os.path.join(self.dbt_project_dir, "dbt_project.yml"), "r") as f:
            self.dbt_project_yml = yaml.load(f, yaml.CLoader)

        # get target path
        if hasattr(self, "target_path"):
            return
        elif "DBT_TARGET_PATH" in os.environ:
            self.target_path = os.path.join(
                self.dbt_project_dir, os.environ["DBT_TARGET_PATH"]
            )
        elif "target-path" in self.dbt_project_yml and tuple(self.version_list[:2]) <= (
            1,
            6,
        ):  # target-path deprecated in 1.7
            self.target_path = os.path.join(
                self.dbt_project_dir, self.dbt_project_yml["target-path"]
            )
        else:
            self.target_path = os.path.join(self.dbt_project_dir, "target")

    def dbt_runner(self, command: list[str]) -> tuple[str, str, bool]:
        with cd(self.dbt_project_dir):
            try:
                from dbt.cli.main import dbtRunner, dbtRunnerResult
            except ImportError:
                raise Exception("dbt 1.5+ runner not installed")

            runner = dbtRunner()
            stdout_buffer = StringIO()
            stderr_buffer = StringIO()

            # Save original streams
            orig_stdout = sys.stdout
            orig_stderr = sys.stderr

            # Redirect streams to buffers
            sys.stdout = stdout_buffer
            sys.stderr = stderr_buffer

            current_dbt_profiles_dir = os.environ.get("DBT_PROFILES_DIR")
            os.environ["DBT_PROFILES_DIR"] = self.dbt_profiles_dir
            result: dbtRunnerResult = runner.invoke(
                command, send_anonymous_usage_stats=False
            )
            del os.environ["DBT_PROFILES_DIR"]
            if current_dbt_profiles_dir:
                os.environ["DBT_PROFILES_DIR"] = current_dbt_profiles_dir

            captured_stdout = stdout_buffer.getvalue()
            captured_stderr = stderr_buffer.getvalue()

            # Restore original streams
            sys.stdout = orig_stdout
            sys.stderr = orig_stderr

            # Return captured output along with function's result
            return captured_stdout, captured_stderr, result.success

    def dbt_cli(self, command: list[str]) -> tuple[str, str, bool]:
        result = subprocess.run(
            [
                "dbtenv",
                "--profiles-dir",
                self.dbt_profiles_dir,
                "execute",
                "--dbt",
                MAX_DIALECT_VERSION[self.dialect.value][self.version.value],
                "--",
                *command,
            ],
            capture_output=True,
            env={
                **os.environ.copy(),
                **{"DBT_PROFILES_DIR": self.dbt_profiles_dir},
                **{"DBT_TARGET_PATH": self.target_path},
                **self.env_vars,
            },
            text=True,
            cwd=self.dbt_project_dir,
        )
        success = not DBTError.has_dbt_error(result.stdout)

        return result.stdout, result.stderr, success

    def run_dbt_command(
        self,
        command: list[str],
        cli_args: list[str] | None = None,
        write_json: bool = False,
        dbt_cache: bool = False,
        force_terminal: bool = True,  # for our purposes, terminal execution is preferred due to version isolation
    ) -> tuple[str, str, bool]:
        if cli_args is None:
            cli_args = []

        full_command = []
        if not write_json:
            full_command.append("--no-write-json")

        full_command.append("--no-use-colors")

        if not self.dbt1_5:
            full_command.append("--no-anonymous-usage-stats")

        if dbt_cache:
            pass
        elif self.dbt1_5:
            full_command.append("--no-populate-cache")
        else:
            full_command.append("--cache-selected-only")

        full_command += command + cli_args

        if self.dbt1_5 and not force_terminal:
            stdout, stderr, success = self.dbt_runner(full_command)
        else:
            stdout, stderr, success = self.dbt_cli(full_command)

        print(stdout, stderr, success)

        if len(stderr) > 0:
            success = False
        return stdout, stderr, success

    def dbt_parse(self) -> tuple[str, str, bool]:
        if self.dbt1_5:
            return self.run_dbt_command(["parse"], write_json=True)
        return self.run_dbt_command(["parse", "--write-manifest"], write_json=True)

    def dbt_compile(
        self,
        node_ids: list[str] | None = None,
        write_json: bool = True,
        dbt_cache: bool = False,
        predecessor_depth: int | None = 0,
        successor_depth: int | None = 0,
        update_manifest=False,
        models_only: bool = False,
    ) -> tuple[str, str, bool]:
        if not node_ids:
            command = ["compile", "-f"]
            node_ids = []
        else:
            command = ["compile", "-f", "-s"]
            full_node_list = self.get_ancestors_and_descendants(
                node_ids, predecessor_depth, successor_depth
            )
            command += [v.split(".")[-1] for v in full_node_list]

        if models_only:
            command.append("--exclude")
            command.append("test_type:singular")
            command.append("test_type:generic")

        # don't exclude node if explicitly provided
        adj_exclusions = (
            list(
                set(self.compile_exclusions)
                - set([id.split(".")[-1] for id in node_ids])
            )
            if self.compile_exclusions
            else []
        )
        if adj_exclusions:
            if "--exclude" not in command:
                command.append("--exclude")
            command.extend(self.compile_exclusions)

        stdout, stderr, success = self.run_dbt_command(
            command, write_json=write_json, dbt_cache=dbt_cache
        )
        if update_manifest:
            self.mount_manifest(force_read=True)
        return stdout, stderr, success

    def dbt_docs_generate(self, compile=False) -> tuple[str, str, bool]:
        if compile:
            stdout, stderr, success = self.run_dbt_command(
                ["docs", "generate"], write_json=True, dbt_cache=True
            )
            return stdout, stderr, success

        stdout, stderr, success = self.run_dbt_command(
            ["docs", "generate", "--no-compile"], write_json=True, dbt_cache=True
        )
        return stdout, stderr, success

    def get_compiled_sql(
        self,
        node_id: str,
        force_read: bool = False,
        errors: list[VinylError] = [],
        compile_if_not_found: bool = True,
    ) -> tuple[str | None, list[VinylError]]:
        self.mount_manifest(force_read=force_read)
        if node_id in self.manifest["nodes"]:
            node = self.manifest["nodes"][node_id]
        elif (
            node_id in self.manifest["disabled"]
            and len(self.manifest["disabled"][node_id]) > 0
        ):
            node = self.manifest["disabled"][node_id][0]

        else:
            node = None

        if node is None:
            errors.extend(
                [
                    VinylError(
                        node_id=node_id,
                        type=VinylErrorType.FILE_NOT_FOUND,
                        msg=f"Node not found: {node_id.split('.')[-1]}",
                        traceback=None,
                        dialect=self.dialect,
                    )
                ]
            )
            return None, errors

        compiled_sql_abs_location = os.path.join(
            self.compiled_sql_path,
            node_id.split(".")[1],
            node["original_file_path"],
        )

        try:
            if not os.path.exists(compiled_sql_abs_location) and compile_if_not_found:
                stdout, stderr, success = self.dbt_compile(
                    [node_id], write_json=True, dbt_cache=False, update_manifest=False
                )
            with open(compiled_sql_abs_location, "r") as f:
                out = f.read()
            return out, errors

        except FileNotFoundError:
            msg = f"Compiled SQL not found for {node_id.split('.')[-1]}"
            if locals().get("stdout", None):
                msg += f". DBT output: {stdout}"
            errors.extend(
                [
                    VinylError(
                        node_id=node_id,
                        type=VinylErrorType.FILE_NOT_FOUND,
                        msg=msg,
                        traceback=traceback.format_exc(),
                        dialect=self.dialect,
                    )
                ]
            )
            return None, errors

    def get_profiles_dir(self):
        if "DBT_PROFILES_DIR" in os.environ:
            return os.environ["DBT_PROFILES_DIR"]

        in_directory_path = os.path.join(self.dbt_project_dir, "profiles.yml")
        if os.path.exists(in_directory_path):
            return os.path.dirname(in_directory_path)

        in_home_path = os.path.join(os.path.expanduser("~"), ".dbt", "profiles.yml")
        if os.path.exists(in_home_path):
            return os.path.dirname(in_home_path)

        raise Exception("Could not find profiles.yml file")

    def get_profile_credentials(self, via_dbt=True):
        dbt_profiles_dir = self.get_profiles_dir()
        profile_name = self.dbt_project_yml["profile"]
        with cd(self.dbt_project_dir):
            if via_dbt:
                # import relevant dbt modules
                from dbt.config import RuntimeConfig
                from dbt.flags import set_from_args

                # generate bq connection object
                _args = DBTArgs(
                    dbt_profiles_dir, self.dbt_project_dir, None, None, {}, None
                )
                set_from_args(_args.to_namespace(), _args)
                dbt_project, dbt_profile = RuntimeConfig.collect_parts(_args)
                _profile = dbt_profile
                _config = RuntimeConfig.from_parts(dbt_project, dbt_profile, _args)
                credentials = _config.credentials
            else:
                with open(f"{dbt_profiles_dir}/profiles.yml", "r") as f:
                    contents = yaml.load(f, yaml.CLoader)
                target_name = contents[profile_name]["target"]
                target = contents[profile_name]["outputs"][target_name]
                dataclsname = self.dialect.get_dbt_credentials_dataclass_name()
                fields = [
                    ("method", str | None),
                    ("database", str | None),
                    ("schema", str | None),
                    ("username", str | None),
                    ("password", str | None),
                    ("host", str | None),
                    ("port", int | None),
                    ("keyfile", str | None),
                    ("keyfile_json", str | None),
                    ("path", str | None),
                ]
                datacls = make_dataclass(dataclsname, fields)
                credentials = datacls(
                    method=target.get("method", None),
                    database=target.get("database", None),
                    schema=target.get(
                        "schema", target.get("dataset", None)
                    ),  # bigquery allows for dataset
                    username=target.get("user", None),
                    password=target.get("password", None),
                    host=target.get("host", None),
                    port=target.get("port", None),
                    keyfile=target.get("keyfile", None),
                    keyfile_json=target.get("keyfile_json", None),
                    path=target.get("path", None),
                )
        return credentials

    def get_relation_name(
        self,
        node_name: str,
    ) -> str | None:
        node_header = node_name.split(".")[0]
        if node_header == "source":
            manifest_key = "sources"
        elif (
            node_header == "model" or node_header == "snapshot" or node_header == "seed"
        ):
            manifest_key = "nodes"
        elif node_header in ["metric", "test"]:
            # these nodes aren't materialized
            return None
        else:
            raise Exception(f"Node type not recognized: {node_header}")
        node = self.manifest[manifest_key][node_name]
        if node.get("relation_name", None) is not None:
            relation_name = node["relation_name"]
        else:
            if (
                node.get("database", None) is not None
                and node.get("schema", None) is not None
                and node.get("name", None) is not None
            ):
                if node.get("alias", None) is not None:
                    tbl_name = node["alias"]
                else:
                    tbl_name = node["name"]
                if self.dialect in ["bigquery", "databricks"]:
                    relation_name = (
                        f"`{node['database']}`.`{node['schema']}`.`{tbl_name}`"
                    )
                else:
                    relation_name = f"{node['database']}.{node['schema']}.{tbl_name}"
            else:
                return None
        return relation_name


def install_dbt_version(
    dialect: DBTDialect,
    major_version: DBTVersion,
):
    need_to_install = False
    try:
        from dbt import version

        version_diff = (
            version.__version__.split(".")[:2] != major_version.value.split(".")[:2]
        )
        if version_diff:
            need_to_install = True

    except ImportError:
        need_to_install = True

    if need_to_install:
        result = subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                f"dbt-{dialect.value}~={major_version.value}.0",
            ],
            capture_output=True,
            text=True,
        )
    if not need_to_install or result.returncode == 0:
        return True
    return False


def run_adjusted_replace(base_pattern_list, replacement, contents):
    pattern = r"\s*".join(base_pattern_list)
    adjusted_pattern = pattern
    return re.sub(adjusted_pattern, replacement, contents)


def replace_sources_and_refs(project: DBTProject, contents):
    try:
        project.mount_manifest()
        project.get_project_yml_files()

        extracted = py_extract_from_source(contents)
        new_contents = contents
        manifest_info = project.manifest
        nodes = manifest_info["nodes"]
        sources = manifest_info["sources"]
        proj_name = project.dbt_project_yml["name"]
        macro_dir = os.path.join(project.dbt_project_dir, "macros")
        # check if user has custom ref macro
        if not file_exists_in_directory("ref.sql", macro_dir):
            for ref in extracted["refs"]:
                model_name = ref["name"]
                package_name = ref.get("package", proj_name)
                base_pattern_list = [
                    r"{{",
                    r"ref",
                    r"\(",
                    r"[\"']",
                    re.escape(model_name),
                    r"[\"']",
                    r"\)",
                    r"}}",
                ]
                node_id = f"model.{package_name}.{model_name}"
                if node_id in nodes:
                    replacement = project.get_relation_name(node_id)
                    new_contents = run_adjusted_replace(
                        base_pattern_list, replacement, new_contents
                    )

        # check if user has custom source macro
        if not file_exists_in_directory("source.sql", macro_dir):
            for source in extracted["sources"]:
                base_pattern_list = [
                    r"{{",
                    r"source",
                    r"\(",
                    r"[\"']",
                    re.escape(source[0]),
                    r"[\"']",
                    r",",
                    r"[\"']",
                    re.escape(source[1]),
                    r"[\"']",
                    r"\)",
                    r"}}",
                ]
                source_id = f"source.{proj_name}.{source[0]}.{source[1]}"
                if source_id in sources:
                    replacement = project.get_relation_name(source_id)
                    new_contents = run_adjusted_replace(
                        base_pattern_list, replacement, new_contents
                    )
        return new_contents

    except ExtractionError:
        return contents


def replace_configs_and_this(
    project: DBTProject, contents, model_node_string: str | None = None
):
    try:
        project.mount_manifest()
        project.get_project_yml_files()

        new_contents = contents
        pattern = r"{{\s*config\s*\([\s\S]*?\)\s*}}"
        new_contents = re.sub(
            pattern, lambda x: "", new_contents, flags=re.MULTILINE | re.DOTALL
        )
        if model_node_string is not None:
            base_pattern_list = [
                r"{{",
                r"this",
                r"}}",
            ]
            replacement = project.get_relation_name(model_node_string)
            new_contents = run_adjusted_replace(
                base_pattern_list, replacement, new_contents
            )
        return new_contents

    except ExtractionError:
        return contents


def fast_compile(project: DBTProject, dbt_sql: str):
    contents = replace_sources_and_refs(project, dbt_sql)
    contents = replace_configs_and_this(project, contents)
    return contents



# helper to make connect imports easier
from vinyl.lib.connect import (  # noqa: F401
    BigQueryConnector,
    DatabaseFileConnector,
    DBTConnectorFull,
    FileConnector,
    PostgresConnector,
    SnowflakeConnector,
)



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_json  # noqa F401


@source(resource=local_json)
class TestJson:
    _table = "test_json"
    _unique_name = "local_json.TestJson"
    _path = "data/jsons/test_json.json"
    _row_count = 10
    _col_replace = {}

    id: t.Int64(nullable=True)
    name: t.String(nullable=True)
    age: t.Int64(nullable=True)
    email: t.String(nullable=True)
    city: t.String(nullable=True)
    country: t.String(nullable=True)
    occupation: t.String(nullable=True)
    salary: t.Int64(nullable=True)
    experience_years: t.Int64(nullable=True)
    remote: t.Boolean(nullable=True)



from __project_name__.sources.local_filesystem.store_num_transactions import (
    StoreNumTransactions,
)
from __project_name__.sources.local_filesystem.stores import Stores
from vinyl import M, T, join, metric, model


@model(deps=[Stores, StoreNumTransactions])
def top_stores(stores: T, txns: T) -> T:
    j = join(stores, txns, on=["store_nbr"])
    j.aggregate(
        cols={"num_transactions": j.transactions.sum()},
        by=[j.store_nbr, j.city, j.state],
    )
    j.sort(by=-j.num_transactions)
    return j


@model(deps=[StoreNumTransactions, Stores])
def store_txns(txns: T, stores: T) -> T:
    table = join(stores, txns, on=[stores.store_nbr == txns.store_nbr])
    return table


@metric(deps=[store_txns])
def sales_metrics(table: T) -> M:
    metric_store = table.metric(
        by=[table.store_nbr, table.city, table.state],
        ts=table.date.cast("timestamp"),
        cols={
            "total_txns": table.transactions.sum(),
            "average_txns": table.transactions.mean(),
        },
    )
    return metric_store



import os
import subprocess
import tempfile
from functools import lru_cache
from pathlib import Path

from dotenv import dotenv_values


@lru_cache()
def load_dotenv(dotenv_path: str | Path | None = None):
    if dotenv_path is None:
        dotenv_path = ".env"
    env_values = dotenv_values(dotenv_path)

    # Check if any op env variables are loaded
    op_variables = [k for k, v in env_values.items() if v.startswith("op://")]
    if op_variables:
        # Fetch the values, will log you in if you need
        command = f"op run --env-file={dotenv_path} --no-masking -- printenv"
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, check=True
        )
        # get the values into python
        with tempfile.NamedTemporaryFile("w+", delete=True) as f:
            f.write(result.stdout)
            f.flush()
            env_values_new = dotenv_values(f.name)
        for k, v in env_values_new.items():
            if k in op_variables:
                env_values[k] = v

    # Set the environment variables
    for k, v in env_values.items():
        os.environ[k] = v



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_parquet  # noqa F401


@source(resource=local_parquet)
class HousePrices:
    _table = "house_prices"
    _unique_name = "local_parquet.HousePrices"
    _path = "data/parquets/house_prices.parquet"
    _row_count = 545
    _col_replace = {}

    price: t.Int64(nullable=True)
    area: t.Int64(nullable=True)
    bedrooms: t.Int64(nullable=True)
    bathrooms: t.Int64(nullable=True)
    stories: t.Int64(nullable=True)
    mainroad: t.String(nullable=True)
    guestroom: t.String(nullable=True)
    basement: t.String(nullable=True)
    hotwaterheating: t.String(nullable=True)
    airconditioning: t.String(nullable=True)
    parking: t.Int64(nullable=True)
    prefarea: t.String(nullable=True)
    furnishingstatus: t.String(nullable=True)



import pytest

from internal_project.sources.local_filesystem.store_num_transactions import (
    StoreNumTransactions,
)
from internal_project.sources.local_filesystem.stores import Stores
from vinyl.lib.set import join


def test_two_way_auto_join_with_match():
    a = Stores()
    b = StoreNumTransactions()
    y = a.select({"cit": a.city, "nbr": a.store_nbr})
    z = b.select({"nbr2": b.store_nbr, "t": b.transactions})

    j = join(y, z)
    j2 = join(z, y)
    j3 = y * z
    j4 = z * y

    assert all(len(t._conn_replace.items()) == 2 for t in [j, j2, j3, j4])

    assert all(
        [
            t._reproducible_hash() == "3f7641d3d68c717ab276776599ca72d2"
            for t in [j, j2, j3, j4]
        ]
    )


def test_two_way_auto_join_without_match():
    a = Stores()
    b = StoreNumTransactions()
    y = a.select({"cit": a.city, "nbr": a.store_nbr})
    z = b.select({"t": b.transactions})

    # automatic joins with the join function don't allow cross joins by default
    with pytest.raises(ValueError):
        join(y, z)
        j2 = join(z, y)

    # automatic joins with the multiply symbol allow for up to one cross join, these have different hashes because cross join order is based on input order by default
    j1 = join(y, z, auto_allow_cross_join=True)
    j2 = join(z, y, auto_allow_cross_join=True)
    j3 = y * z
    j4 = z * y

    assert all(len(t._conn_replace.items()) == 2 for t in [j1, j2, j3, j4])

    assert all(
        [t._reproducible_hash() == "a28237374e96fa9c2e6ee31bab97424a" for t in [j1, j3]]
    )
    assert all(
        [t._reproducible_hash() == "dcba46d0f9dae45d48b5fa0b01fa7ca4" for t in [j2, j4]]
    )
